@@ -1,5 +1,4 @@
 from typing import (
-    Any,
     Dict,
     List,
     Optional,
@@ -8,7 +7,9 @@ from jsonpointer import resolve_pointer
 import json_merge_patch
 import urllib.request
 
-from .common.jsonschema import map_common_json_schema_fields
+from ..validation import validate_sdf_model
+
+from .jsonschema import map_common_json_schema_fields
 from .utility import (
     initialize_list_field,
     initialize_object_field,
@@ -72,6 +73,7 @@ def resolve_sdf_ref(
             try:
                 with urllib.request.urlopen(resolved_url) as url:
                     retrieved_sdf_model = json.loads(url.read().decode())
+                    validate_sdf_model(retrieved_sdf_model)
                     original = resolve_pointer(retrieved_sdf_model, "/" + pointer)
                     original = resolve_sdf_ref(
                         retrieved_sdf_model, original, resolved_url, sdf_ref_list
@@ -303,7 +305,7 @@ def map_data_qualities(
     )
     map_enum(data_qualities, data_schema, mapped_fields)
 
-    # TODO: Revisit the mapping of these two fields
+    map_unique_items(data_qualities, data_schema, suppress_roundtripping, mapped_fields)
     map_nullable(data_qualities, data_schema, suppress_roundtripping, mapped_fields)
     map_sdf_type(data_qualities, data_schema, suppress_roundtripping, mapped_fields)
 
@@ -353,21 +355,21 @@ def map_readable(sdf_property, wot_property, mapped_fields):
 def map_properties(
     sdf_model, data_qualities, data_schema, suppress_roundtripping: bool, mapped_fields
 ):
-    properties = data_qualities.get("properties")
+    sdf_properties = data_qualities.get("properties")
 
-    if properties is None:
+    if sdf_properties is None:
         return
 
     mapped_fields.append("properties")
 
-    for key, property in properties.items():
-        initialize_object_field(data_schema, "properties")
-        data_schema["properties"][key] = {}
+    for key, sdf_property in sdf_properties.items():
+        wot_properties = initialize_object_field(data_schema, "properties")
+        wot_property = initialize_object_field(wot_properties, key)
         mapped_property_fields: List[str] = []
         map_data_qualities(
             sdf_model,
-            property,
-            data_schema["properties"][key],
+            sdf_property,
+            wot_property,
             suppress_roundtripping,
             mapped_property_fields,
         )
@@ -378,12 +380,13 @@ def map_items(
 ):
     if "items" in data_qualities:
         mapped_fields.append("items")
-        data_schema["items"] = {}
+        sdf_items = data_qualities["items"]
+        wot_items = initialize_object_field(data_schema, "items")
         mapped_item_fields: List[str] = []
         map_data_qualities(
             sdf_model,
-            data_qualities["items"],
-            data_schema["items"],
+            sdf_items,
+            wot_items,
             suppress_roundtripping,
             mapped_item_fields,
         )
@@ -409,6 +412,21 @@ def map_sdf_type(
     )
 
 
+def map_unique_items(
+    data_qualities, data_schema, suppress_roundtripping: bool, mapped_fields: List[str]
+):
+    if suppress_roundtripping:
+        return
+
+    map_field(
+        data_qualities,
+        data_schema,
+        "uniqueItems",
+        "sdf:uniqueItems",
+        mapped_fields=mapped_fields,
+    )
+
+
 def map_nullable(
     data_qualities, data_schema, suppress_roundtripping: bool, mapped_fields
 ):
@@ -446,28 +464,30 @@ def map_action_qualities(
     json_pointer: str,
     suppress_roundtripping: bool,
 ):
-    initialize_object_field(thing_model, "actions")
-    affordance_key = "_".join(prefix_list)
-
-    wot_action: Dict[str, Any] = {}
-    mapped_fields: List[str] = []
-    collect_sdf_required(thing_model, sdf_action, mapped_fields)
-    collect_mapping(thing_model, json_pointer, "actions", affordance_key)
     sdf_action = resolve_sdf_ref(sdf_model, sdf_action, None, [])
 
+    wot_actions = initialize_object_field(thing_model, "actions")
+    action_key = "_".join(prefix_list)
+    wot_action = initialize_object_field(wot_actions, action_key)
+    mapped_fields: List[str] = []
+
+    collect_sdf_required(thing_model, sdf_action, mapped_fields)
+    collect_mapping(thing_model, json_pointer, "actions", action_key)
+
     map_common_qualities(sdf_action, wot_action, suppress_roundtripping, mapped_fields)
 
     data_map_pairs = [("sdfInputData", "input"), ("sdfOutputData", "output")]
 
-    for sdf_field, wot_field in data_map_pairs:
-        if sdf_field in sdf_action:
-            mapped_fields.append(sdf_field)
-            wot_action[wot_field] = {}
+    for sdf_field_name, wot_field_name in data_map_pairs:
+        if sdf_field_name in sdf_action:
+            mapped_fields.append(sdf_field_name)
+            sdf_data_qualities = sdf_action[sdf_field_name]
+            wot_data_schema = initialize_object_field(wot_action, wot_field_name)
             mapped_data_quality_fields: List[str] = []
             map_data_qualities(
                 sdf_model,
-                sdf_action[sdf_field],
-                wot_action[wot_field],
+                sdf_data_qualities,
+                wot_data_schema,
                 suppress_roundtripping,
                 mapped_data_quality_fields,
             )
@@ -485,23 +505,21 @@ def map_action_qualities(
 
     map_additional_fields(wot_action, sdf_action, mapped_fields)
 
-    thing_model["actions"][affordance_key] = wot_action
-
 
 def map_property_qualities(
     sdf_model: Dict,
     thing_model: Dict,
     sdf_property: Dict,
-    affordance_key: str,
+    property_key: str,
     json_pointer: str,
     suppress_roundtripping: bool,
 ):
-    initialize_object_field(thing_model, "properties")
+    wot_properties = initialize_object_field(thing_model, "properties")
+    wot_property = initialize_object_field(wot_properties, property_key)
 
-    wot_property: Dict[str, Any] = {}
     mapped_fields: List[str] = []
     collect_sdf_required(thing_model, sdf_property, mapped_fields)
-    collect_mapping(thing_model, json_pointer, "properties", affordance_key)
+    collect_mapping(thing_model, json_pointer, "properties", property_key)
 
     map_data_qualities(
         sdf_model,
@@ -514,24 +532,26 @@ def map_property_qualities(
 
     map_additional_fields(wot_property, sdf_property, mapped_fields)
 
-    thing_model["properties"][affordance_key] = wot_property
-
 
 # TODO: Find a better name for this function
 def map_sdf_data_qualities(
     sdf_model: Dict,
     thing_model: Dict,
     sdf_data: Dict,
-    affordance_key: str,
+    schema_definition_key: str,
     json_pointer: str,
     suppress_roundtripping: bool,
 ):
-    initialize_object_field(thing_model, "schemaDefinitions")
+    schema_definitions = initialize_object_field(thing_model, "schemaDefinitions")
+    wot_schema_definition = initialize_object_field(
+        schema_definitions, schema_definition_key
+    )
 
-    wot_schema_definition: Dict[str, Any] = {}
     mapped_fields: List[str] = []
     collect_sdf_required(thing_model, sdf_data, mapped_fields)
-    collect_mapping(thing_model, json_pointer, "schemaDefinitions", affordance_key)
+    collect_mapping(
+        thing_model, json_pointer, "schemaDefinitions", schema_definition_key
+    )
 
     map_data_qualities(
         sdf_model,
@@ -548,8 +568,6 @@ def map_sdf_data_qualities(
 
     map_additional_fields(wot_schema_definition, sdf_data, mapped_fields)
 
-    thing_model["schemaDefinitions"][affordance_key] = wot_schema_definition
-
 
 def map_sdf_data(
     sdf_model: Dict,
@@ -654,25 +672,25 @@ def map_event_qualities(
     json_pointer: str,
     suppress_roundtripping: bool,
 ):
-    initialize_object_field(thing_model, "events")
-    affordance_key = "_".join(prefix_list)
+    wot_events = initialize_object_field(thing_model, "events")
+    event_key = "_".join(prefix_list)
+    wot_event = initialize_object_field(wot_events, event_key)
 
-    wot_event: Dict[str, Any] = {}
     mapped_fields: List[str] = []
     collect_sdf_required(thing_model, sdf_event, mapped_fields)
-    collect_mapping(thing_model, json_pointer, "events", affordance_key)
+    collect_mapping(thing_model, json_pointer, "events", event_key)
     sdf_event = resolve_sdf_ref(sdf_model, sdf_event, None, [])
 
     map_common_qualities(sdf_event, wot_event, suppress_roundtripping, mapped_fields)
 
     if "sdfOutputData" in sdf_event:
         mapped_fields.append("sdfOutputData")
-        wot_event["data"] = {}
+        wot_data_schema = initialize_object_field(wot_event, "data")
         mapped_output_data_fields: List[str] = []
         map_data_qualities(
             sdf_model,
             sdf_event["sdfOutputData"],
-            wot_event["data"],
+            wot_data_schema,
             suppress_roundtripping,
             mapped_output_data_fields,
         )
@@ -690,8 +708,6 @@ def map_event_qualities(
 
     map_additional_fields(wot_event, sdf_event, mapped_fields)
 
-    thing_model["events"][affordance_key] = wot_event
-
 
 def collect_sdf_required(
     thing_model: Dict, sdf_definition: Dict, mapped_fields: List[str]
@@ -1009,6 +1025,8 @@ def consolidate_sdf_model(sdf_model: Dict, sdf_mapping_files: List[Dict]):
     for sdf_mapping_file in sdf_mapping_files:
         _apply_mapping_file(sdf_model, sdf_mapping_file)
 
+    validate_sdf_model(sdf_model, framework=True)
+
 
 def convert_sdf_to_wot_tm(
     sdf_model: Dict,
@@ -1018,6 +1036,8 @@ def convert_sdf_to_wot_tm(
     suppress_roundtripping=False,
 ) -> Dict[str, Dict]:
 
+    validate_sdf_model(sdf_model)
+
     if sdf_mapping_files is not None:
         consolidate_sdf_model(sdf_model, sdf_mapping_files)
 
